#include "spawn.h"

#include "..\algorithms.h"

#include "..\dllcore.cpp"
#include "..\datatypes.cpp"
#include "..\algorithms.cpp"
#include "..\player.cpp"
#include "..\commtypes.cpp"


//////// Bot list ////////

_linkedlist <botInfo> botlist;

botInfo *findBot(CALL_HANDLE handle)
{
	_listnode <botInfo> *parse = botlist.head;

	while (parse)
	{
		botInfo *item = parse->item;

		if (item->validate(handle))
			return item;

		parse = parse->next;
	}

	return 0;
}


//////// DLL "import" ////////

void botInfo::tell(BotEvent &event)
{
	if (callback && handle)
	{
		event.handle = handle;
		callback(event);
	}
}

void botInfo::sendReset()
{
	if (*arena && !isNumeric(arena))
	{	// if it's in a private arena
		sendPublic("*scorereset");
	}
}

bool botInfo::sendGreet()
{
	if (!allowZone) return false;

	countdown[1] = zonedelay;
	allowZone = false;

	String s;
	s += "*zone Elimination is starting. Type ?go ";
	s += arena;
	sendPublic(s);

	return true;
}

void botInfo::setfreq(Player *p, int team)
{
	if ((team < 0) || (team > 9999))
		return;

	String s;
	s += "*setfreq ";
	s += team;
	sendPrivate(p, s);

	p->team = team;
}

void botInfo::setship(Player *p, int ship)
{
	if ((ship < SHIP_Warbird) || (ship > SHIP_Shark))
		return;

	String s;
	s += "*setship ";
	s += ship + 1;
	sendPrivate(p, s);

	p->ship = ship;
}

void botInfo::spec(Player *p)
{
	sendPrivate(p, "*spec");
	sendPrivate(p, "*spec");

	p->ship = SHIP_Spectator;
}

void botInfo::findWinner()
{
	_listnode <Player> *parse = playerlist->head;

	Player *winner = 0;
	int total = 0;

	while (parse)
	{
		Player *p = parse->item;

		if (p->tags[0])
		{
			++total;
			winner = p;

			if (game == GAME_Team)
			{
				if (total > TEAM_MAX)
					return;
			}
			else
			{
				if (total > 1)
					return;
			}
		}

		parse = parse->next;
	}

	if (total == 0)
	{
		sendPublic(SND_VictoryBell, "*arena No winner.");
	}
	else
	{
		String s;
		s += "*arena Game over. Winner: ";

		if (game == GAME_Team)
		{
			s += "frequency ";
			s += winner->team;
		}
		else
		{
			s += winner->name;
		}

		sendPublic(SND_VictoryBell, s);
	}

	if (getIngame() >= 2)
	{
		state = ELIM_Announcing;
		countdown[0] = 5;
	}
	else
	{
		state = ELIM_EnterGame;
		countdown[0] = 3;
	}

	if ((game != GAME_Zombies) && (game != GAME_Race))
	{
		sendTeam("You may enter");
		sendPublic("*lock");
	}
}

void botInfo::makeSingles()
{
	static BYTE fmat[10000];
	ZeroMemory(fmat, 10000);

	_listnode <Player> *parse = playerlist->head;

	if (game == GAME_Zombies)
	{
		fmat[ZOMBIE_TEAM] = true;
	}

	while (parse)
	{
		Player *p = parse->item;

		parse = parse->next;

		if (p->ship == SHIP_Spectator)
		{
			p->tags[0] = false;

			continue;
		}

		p->tags[0] = true;

		if (game == GAME_Ship)
		if (p->ship != SHIP_Warbird)
			setship(p, SHIP_Warbird);

		if (game == GAME_Zombies)
		if (p->ship == SHIP_Spider)
			setship(p, SHIP_Warbird);

		int team = p->team;

		if ((team < 0) || (team > 9999))
			continue;

		if (fmat[team])
		{
			for (int i = 50; i < 10000; ++i)
			{
				if (!fmat[i])
				{
					break;
				}
			}

			fmat[i] = true;

			setfreq(p, i);
		}
		else
		{
			fmat[team] = true;
		}
	}
}

int botInfo::getIngame()
{
	_listnode <Player> *parse = playerlist->head;

	int total = 0;

	while (parse)
	{
		Player *p = parse->item;

		if (p->ship != SHIP_Spectator)
			++total;

		parse = parse->next;
	}

	return total;
}

void botInfo::elimTeam(int team)
{
	_listnode <Player> *parse = playerlist->head;

	Player *tp[TEAM_MAX];
	ZeroMemory(tp, sizeof(Player*) * TEAM_MAX);

	int j, i = 0;

	while (parse)
	{
		Player *p = parse->item;

		if (p->team == team)	
		{
			tp[i++] = p;
			if (i >= TEAM_MAX)
				break;
		}

		parse = parse->next;
	}

	if (i >= 1)
	{
		int losses = 0, wins = 0;

		for (j = 0; j < i; ++j)
			if (tp[j])
			{
				losses	+= tp[j]->score.losses;
				wins	+= tp[j]->score.wins;
			}

		if (losses >= deaths)
		{
			for (j = 0; j < i; ++j)
			{
				if (tp[j])
				{
					spec(tp[j]);
					tp[j]->tags[0] = false;
				}
			}

			String s;
			s += "*arena Frequency ";
			s += team;
			s += " eliminated. ";
			s += wins;
			s += " wins; ";
			s += losses;
			s += " losses";
			sendPublic(s);

			findWinner();
		}
	}
}

String botInfo::getStatus()
{
	String s;

	switch (state)
	{
	case ELIM_NotPlaying:
		{
			s += "Elimination has not been started. Send ::!help for commands";
		}
		break;
	case ELIM_VotingGame:
		{
			s += "We are voting on the game. 1=Elim 2=Teams 3=Warbirds 4=Zombies 5=Conquer 6=Race";
		}
		break;
	case ELIM_VotingDeaths:
		{
			s += "We are voting on the number of deaths. (1-9)";
		}
		break;
	case ELIM_GetReady:
		{
			switch (game)
			{
			case GAME_Elim:
				{
					s += "We are playing elimination to ";
					s += deaths;
					if (deaths == 1) s += " death.";
					else			 s += " deaths.";
				}
				break;
			case GAME_Team:
				{
					s += "We are playing team elimination to ";
					s += deaths;
					if (deaths == 1) s += " death.";
					else			 s += " deaths.";
				}
				break;
			case GAME_Ship:
				{
					s += "We are playing warbird-only elimination to ";
					s += deaths;
					if (deaths == 1) s += " death.";
					else			 s += " deaths.";
				}
				break;
			case GAME_Zombies:
				{
					s += "We are playing zombielim to ";
					s += deaths;
					if (deaths == 1) s += " death.";
					else			 s += " deaths.";
				}
				break;
			case GAME_Conquer:
				{
					s += "We are playing conquer.";
				}
				break;
			case GAME_Race:
				{
					s += "We are playing race to ";
					s += deaths;
					if (deaths == 1) s += " kill.";
					else			 s += " kills.";
				}
				break;
			}

			s += " Game will begin in ";
			s += countdown[0];
			if (countdown[0] == 1)	s += " second";
			else				s += " seconds";
		}
		break;
	case ELIM_Playing:
		{
			switch (game)
			{
			case GAME_Elim:
				{
					s += "We are playing elimination to ";
					s += deaths;
					if (deaths == 1) s += " death.";
					else			 s += " deaths.";
				}
				break;
			case GAME_Team:
				{
					s += "We are playing team elimination to ";
					s += deaths;
					if (deaths == 1) s += " death.";
					else			 s += " deaths.";
				}
				break;
			case GAME_Ship:
				{
					s += "We are playing warbird-only elimination to ";
					s += deaths;
					if (deaths == 1) s += " death.";
					else			 s += " deaths.";
				}
				break;
			case GAME_Zombies:
				{
					s += "We are playing zombielim to ";
					s += deaths;
					if (deaths == 1) s += " death.";
					else			 s += " deaths.";
				}
				break;
			case GAME_Conquer:
				{
					s += "We are playing conquer.";
				}
				break;
			case GAME_Race:
				{
					s += "We are playing race to ";
					s += deaths;
					if (deaths == 1) s += " kill.";
					else			 s += " kills.";
				}
				break;
			}

			s += " ";
			s += getIngame();
			s += " players remaining";
		}
		break;
	case ELIM_EnterGame:
		{
			s += "Elimination will restart when 2 players enter. Send ::!help for commands";
		}
		break;
	default:
		{
			s += "I don't know what is going on!";
		}
		break;
	};

	return s;
}

void botInfo::gotEvent(BotEvent &event)
{
	if (CONNECTION_DENIED) return;

	switch (event.code)
	{
//////// Periodic ////////
	case EVENT_Tick:
		{
			for (int i = 0; i < 4; ++i)
				--countdown[i];

			if (countdown[0] == 0)
			switch (state)
			{
			case ELIM_NotPlaying:
				{
					// do nothing
				}
				break;
			case ELIM_Announcing:
				{
					// announce game, wait 20

					sendGreet();
					sendPublic(SND_TrebleBeep, "*arena What do you want to play? 1=Elim 2=Teams 3=Warbirds 4=Zombies 5=Conquer 6=Race");

					state = ELIM_VotingGame;
					countdown[0] = 20;

					for (int i = 0; i < 10; ++i)
						votes[i] = 0;
				}
				break;
			case ELIM_VotingGame:
				{
					// tally votes, 

					int best = 1;

					for (int i = 0; i < 10; ++i)
						if (votes[best] < votes[i]) best = i;

					if ( (best < 1) || (best > 6) )
						game = GAME_Elim;
					else
						game = (Elim_Game)(best - 1);

					switch (game)
					{
					case GAME_Elim:		sendPublic(SND_TrebleBeep, "*arena It is Elim!");			break;
					case GAME_Team:		sendPublic(SND_TrebleBeep, "*arena It is Teams!");			break;
					case GAME_Ship:		sendPublic(SND_TrebleBeep, "*arena It is Warbirds-only!");	break;
					case GAME_Zombies:	sendPublic(SND_TrebleBeep, "*arena It is Zombies!");		break;
					case GAME_Conquer:	sendPublic(SND_TrebleBeep, "*arena It is Conquer!");		break;
					case GAME_Race:		sendPublic(SND_TrebleBeep, "*arena It is Race!");			break;
					}

					switch (game)
					{
					case GAME_Elim:
					case GAME_Team:
					case GAME_Ship:
					case GAME_Zombies:
						{
							sendPublic("*arena Elimination at how many deaths? (1-9)");

							state = ELIM_VotingDeaths;
							countdown[0] = 20;

							for (int i = 0; i < 10; ++i)
								votes[i] = 0;
						}
						break;
					case GAME_Race:
						{
							sendPublic("*arena Victory at how many kills? (1-9)");

							state = ELIM_VotingDeaths;
							countdown[0] = 20;

							for (int i = 0; i < 10; ++i)
								votes[i] = 0;
						}
						break;
					case GAME_Conquer:
						{
							sendPublic("*arena Rules: All on own freqs. Die and join your killer. Owner of last team wins.");
							sendPublic("*arena Get ready. Game will begin in 10 seconds...");

							state = ELIM_GetReady;
							countdown[0] = 10;
						}
						break;
					}
				}
				break;
			case ELIM_VotingDeaths:
				{
					// tally votes, wait 10

					int best = 2;

					for (int i = 0; i < 10; ++i)
						if (votes[best] < votes[i]) best = i;

					deaths = best;

					String s;

					switch (game)
					{
					case GAME_Elim:
						{
							s += "*arena Rules: All on own freqs. Die ";
							s += deaths;
							s += " times and be eliminated. Sole survivor wins.";
							sendPublic(s);
						}
						break;
					case GAME_Team:
						{
							s += "*arena Rules: All on teams of 2. Combined deaths of ";
							s += deaths;
							s += " will eliminate both players. Last team to survive wins.";
							sendPublic(s);
						}
						break;
					case GAME_Ship:
						{
							s += "*arena Rules: All on own freqs. Use warbirds only. Die ";
							s += deaths;
							s += " times and be eliminated. Sole survivor wins.";
							sendPublic(s);
						}
						break;
					case GAME_Zombies:
						{
							s += "*arena Rules: All on own freqs. Die ";
							s += deaths;
							s += " times to get turned into a zombie. Sole human survivor wins.";
							sendPublic(s);
						}
						break;
					case GAME_Race:
						{
							s += "*arena Rules: All on own freqs. Kill ";
							s += deaths;
							s += " times to win.";
							sendPublic(s);
						}
						break;
					}

					sendPublic(SND_TrebleBeep, "*arena Get ready. Game will begin in 10 seconds...");

					state = ELIM_GetReady;
					countdown[0] = 10;
				}
				break;
			case ELIM_GetReady:
				{
					// gogogo or cancel
					int minimum_players = (game == GAME_Team) ? TEAM_MAX * 2 : 2;

					if (getIngame() >= minimum_players)
					{
						if ((game != GAME_Zombies) && (game != GAME_Race))
							sendPublic("*lock");
						sendReset();
						sendPublic(SND_Goal, "*arena Go! Go! Go!");

						if (game != GAME_Team)
						{
							makeSingles();
						}
						else
						{
							_listnode <Player> *parse = playerlist->head;
							int count = 0, freq = 0;

							while (parse)
							{
								Player *p = parse->item;
								parse = parse->next;

								if (p->ship == SHIP_Spectator)
								{
									p->tags[0] = false;

									continue;
								}

								p->tags[0] = true;

								if (p->team != freq)
									setfreq(p, freq);

								++count;
								if (count == TEAM_MAX)
								{
									count = 0;
									++freq;
								}
							}
						}

						state = ELIM_Playing;
					}
					else
					{
						sendPublic(SND_CantLogIn, "*arena Aborted. Elimination will restart when 2 or more players enter.");

						state = ELIM_EnterGame;
						countdown[0] = 3;
					}
				}
				break;
			case ELIM_Playing:
				{
					// do nothing
				}
				break;
			case ELIM_EnterGame:
				{
					if (getIngame() >= 2)
					{
						state = ELIM_Announcing;
						countdown[0] = 1;
					}
					else
					{
						countdown[0] = 3;
					}
				}
				break;
			};

			if (countdown[1] == 0)
			{	// 2 minutes are up
				allowZone = true;
			}
		}
		break;
//////// Arena ////////
	case EVENT_ArenaEnter:
		{
			arena = (char*)event.p[0];
			me = (Player*)event.p[1];	// if(me) {/*we are in the arena*/}
		}
		break;
	case EVENT_ArenaSettings:
		{
			settings = (arenaSettings*)event.p[0];
		}
		break;
	case EVENT_ArenaLeave:
		{
			me = 0;
			state = ELIM_NotPlaying;
		}
		break;
	case EVENT_ArenaListEntry:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
	case EVENT_ArenaListEnd:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
//////// Flag ////////
	case EVENT_FlagGrab:
		{
			Player *p = (Player*)event.p[0];
			Flag *f = (Flag*)event.p[1];
		}
		break;
	case EVENT_FlagDrop:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_FlagMove:
		{
			Flag *f = (Flag*)event.p[0];
		}
		break;
	case EVENT_FlagVictory:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_FlagReward:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// Timed ////////
	case EVENT_TimedGameOver:
		{
			Player *p[5];
			for (int i = 0; i < 5; i++)
				p[i] = (Player*)event.p[i];
		}
		break;
//////// Soccer ////////
	case EVENT_SoccerGoal:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_BallMove:
		{
			PBall *ball = (PBall*)event.p[0];
		}
		break;
//////// File ////////
	case EVENT_File:
		{
			char *name = (char*)event.p[0];
		}
		break;
//////// Player ////////
	case EVENT_PlayerEntering:
		{
			Player *p = (Player*)event.p[0];

			p->tags[0] = false;

			if (state != ELIM_NotPlaying)
			{
				/* Elim logic */
				sendPrivate(p, getStatus());
			}
			else
			{
				/* AutoSpec logic */
				if (autospec_enabled)
				if (p->ship != SHIP_Spectator)
				if (p->score.losses >= autospec_deaths)
				{
					spec(p);
				}
			}

			/* Elim logic */
			if (state == ELIM_Playing)
			if (p->ship != SHIP_Spectator)
			{
				switch (game)
				{
				case GAME_Elim:
				case GAME_Team:
				case GAME_Ship:
				case GAME_Conquer:
						spec(p);
					break;
				case GAME_Zombies:
						setfreq(p, ZOMBIE_TEAM);
						setship(p, SHIP_Spider);
				case GAME_Race:
						sendPrivate(p, "*scorereset");
				}
			}
		}
		break;
	case EVENT_PlayerMove:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerDeath:
		{
			Player *p = (Player*)event.p[0],
				   *k = (Player*)event.p[1];

			if (p->ship == SHIP_Spectator)
				break;

			/* AutoSpec logic */
			if (state == ELIM_NotPlaying)
			{
				if (autospec_enabled)
				if (p->score.losses >= autospec_deaths)
				{
					String s;
					s += "*arena";
					s += p->name;
					s += " eliminated. ";
					s += p->score.wins;
					s += " wins; ";
					s += p->score.losses;
					s += " losses";
					sendPublic(s);
					spec(p);
				}
			}

			/* Elim logic */
			if (state == ELIM_Playing)
			{
				switch (game)
				{
				case GAME_Elim:
				case GAME_Ship:
					{
						if (p->score.losses >= deaths)
						{
							String s;
							s += "*arena";
							s += p->name;
							s += " eliminated. ";
							s += p->score.wins;
							s += " wins; ";
							s += p->score.losses;
							s += " losses";
							sendPublic(s);
							spec(p);
							p->tags[0] = false;

							findWinner();
						}
					}
					break;
				case GAME_Team:
					{
						elimTeam(p->team);
					}
					break;
				case GAME_Zombies:
					{
						if (p->tags[0])
						if (p->score.losses >= deaths)
						{
							String s;
							s += "*arena";
							s += p->name;
							s += " zombified. ";
							s += p->score.wins;
							s += " wins; ";
							s += p->score.losses;
							s += " losses";
							sendPublic(s);
							setfreq(p, ZOMBIE_TEAM);
							setship(p, SHIP_Spider);
							p->tags[0] = false;

							findWinner();
						}
					}
					break;
				case GAME_Conquer:
					{
						int team = p->team;

						setfreq(p, k->team);

						if (p->tags[0])
						{	// he's a team owner, move other players on team also
							_listnode <Player> *parse = playerlist->head;

							while (parse)
							{
								Player *pp = parse->item;

								if (pp != p)
								if (pp->team == team)
								{
									setfreq(pp, k->team);
								}

								parse = parse->next;
							}

							p->tags[0] = false;

							findWinner();
						}
					}
					break;
				case GAME_Race:
					{
						if (k->score.wins >= deaths)
						{
							_listnode <Player> *parse = playerlist->head;

							while (parse)
							{
								Player *pp = parse->item;

								if (pp != k)
								{
									p->tags[0] = false;
								}

								parse = parse->next;
							}

							findWinner();
						}
					}
					break;
				}
			}
		}
		break;
	case EVENT_PlayerScore:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerPrize:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerShip:
		{
			Player *p = (Player*)event.p[0];

			if (p->ship == SHIP_Spectator)
				break;

			/* AutoSpec logic */
			if (state == ELIM_NotPlaying)
			{
				if (autospec_enabled)
				if (p->score.losses >= autospec_deaths)
				{
					String s;
					s += "*arena";
					s += p->name;
					s += " eliminated. ";
					s += p->score.wins;
					s += " wins; ";
					s += p->score.losses;
					s += " losses";
					sendPublic(s);
					spec(p);
				}
			}

			/* Elim logic */
			if (state == ELIM_Playing)
			{
				switch (game)
				{
				case GAME_Ship:
					if (!p->tags[0])
						spec(p);
					else
						if (p->ship != SHIP_Warbird)
							setship(p, SHIP_Warbird);
					break;
				case GAME_Elim:
				case GAME_Team:
				case GAME_Conquer:
					if (!p->tags[0])
						spec(p);
					break;
				case GAME_Zombies:
					if (!p->tags[0])
					{
						if (p->ship != SHIP_Spider)
							setship(p, SHIP_Spider);
						if (p->team != ZOMBIE_TEAM)
							setfreq(p, ZOMBIE_TEAM);
					}
				}
			}
		}
		break;
	case EVENT_PlayerSpec:
		{
			Player *p = (Player*)event.p[0];

			if (state == ELIM_Playing)
			{
				switch (game)
				{
				case GAME_Elim:
				case GAME_Team:
				case GAME_Ship:
				case GAME_Zombies:
					p->tags[0] = false;
					findWinner();
					break;
				case GAME_Conquer:
					if (p->tags[0])
					{
						_listnode <Player> *parse = playerlist->head;

						while (parse)
						{
							Player *pp = parse->item;

							if (pp != p)
							if (pp->team == p->team)
							{
								spec(pp);
							}

							parse = parse->next;
						}

						p->tags[0] = false;
						findWinner();
					}
					break;
				}
			}
		}
		break;
	case EVENT_PlayerTeam:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldteam = (Uint16)event.p[1];

			if (p->ship == SHIP_Spectator)
				break;

			if (state == ELIM_Playing)
			if (oldteam != p->team)
				setfreq(p, oldteam);
		}
		break;
	case EVENT_PlayerLeaving:
		{
			Player *p = (Player*)event.p[0];

			if (p->ship == SHIP_Spectator)
				break;

			if (state == ELIM_Playing)
			{
				switch (game)
				{
				case GAME_Elim:
				case GAME_Team:
				case GAME_Ship:
				case GAME_Zombies:
					findWinner();
					break;
				case GAME_Conquer:
					if (p->tags[0])
					{
						_listnode <Player> *parse = playerlist->head;

						while (parse)
						{
							Player *pp = parse->item;

							if (pp != p)
							if (pp->team == p->team)
							{
								spec(pp);
							}

							parse = parse->next;
						}

						p->tags[0] = false;
						findWinner();
					}
					break;
				}
			}
		}
		break;
//////// Turret ////////
	case EVENT_CreateTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
	case EVENT_DeleteTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
//////// Chat ////////
	case EVENT_Chat:
		{
			int type  = *(int*)&event.p[0];
			int sound = *(int*)&event.p[1];
			Player *p = (Player*)event.p[2];
			char *msg = (char*)event.p[3];

			switch (type)
			{
			case MSG_Arena:
				{
					if (state == ELIM_Playing)
					{
						if (CMPSTR(msg, "Arena UNLOCKED"))
						{
							sendPublic("*lock");
						}
					}
					else
					{
						if (CMPSTR(msg, "Arena LOCKED"))
						{
							sendPublic("*lock");
						}
					}
				}
				break;
			case MSG_PublicMacro:
			case MSG_Public:
			case MSG_Team:
			case MSG_TeamPrivate:
			case MSG_Private:			if (!p) break;
				{
					if (isNumeric(msg))
					{
						int vote = getInteger(msg, 10);
						if (vote < 1 || vote > 9) break;

						++votes[vote];
					}
				}
				break;
			case MSG_PlayerWarning:		if (!p) break;
				{
				}
				break;
			case MSG_RemotePrivate:
				{
				}
				break;
			case MSG_ServerError:
				{
				}
				break;
			case MSG_Channel:
				{
				}
				break;
			};
		}
		break;
	case EVENT_LocalCommand:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotCommand(p, c);
		}
		break;
	case EVENT_LocalHelp:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotHelp(p, c);
		}
		break;
	case EVENT_RemoteCommand:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemote(p, c, o);
		}
		break;
	case EVENT_RemoteHelp:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemoteHelp(p, c, o);
		}
		break;
//////// Containment ////////
	case EVENT_Init:
		{
			int major = HIWORD(*(int*)&event.p[0]);
			int minor = LOWORD(*(int*)&event.p[0]);
			callback = (CALL_COMMAND)event.p[1];
			playerlist = (CALL_PLIST)event.p[2];
			flaglist = (CALL_FLIST)event.p[3];

			if (major > CORE_MAJOR_VERSION)
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin is out of date."));

				CONNECTION_DENIED = true;

				return;
			}
			else if ((major < CORE_MAJOR_VERSION) || (minor < CORE_MINOR_VERSION))
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin requires the latest version of MERVBot."));

				CONNECTION_DENIED = true;

				return;
			}
			else
				tell(makeEcho("DLL plugin connected."));
		}
		break;
	case EVENT_Term:
		{
			tell(makeEcho("DLL plugin disconnected."));
		}
		break;
	default:
		{
			String s;
			s += "Unknown message from bot (";
			s += event.code;
			s += ")  Version mismatch?";
			tell(makeEcho(s));
		}
		break;
	};
}


//////// DLL export ////////

_declspec(dllexport) void __stdcall talk(BotEvent &event)
{
	botInfo *bot;

	bot = findBot(event.handle);

	if (!bot)
	{
		bot = new botInfo(event.handle);
		botlist.append(bot);
	}

	if (bot) bot->gotEvent(event);
}


//////// Chatter ////////

void botInfo::sendPrivate(Player *player, BYTE snd, char *msg)
{
	tell(makeSay(MSG_Private, snd, player->ident, msg));
}

void botInfo::sendPrivate(Player *player, char *msg)
{
	tell(makeSay(MSG_Private, 0, player->ident, msg));
}

void botInfo::sendTeam(char *msg)
{
	tell(makeSay(MSG_Team, 0, 0, msg));
}

void botInfo::sendTeam(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Team, snd, 0, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, 0, team, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, BYTE snd, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, snd, team, msg));
}

void botInfo::sendPublic(char *msg)
{
	tell(makeSay(MSG_Public, 0, 0, msg));
}

void botInfo::sendPublic(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Public, snd, 0, msg));
}

void botInfo::sendPublicMacro(char *msg)
{
	tell(makeSay(MSG_PublicMacro, 0, 0, msg));
}

void botInfo::sendPublicMacro(BYTE snd, char *msg)
{
	tell(makeSay(MSG_PublicMacro, snd, 0, msg));
}

void botInfo::sendChannel(char *msg)
{
	tell(makeSay(MSG_Channel, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *msg)
{
	tell(makeSay(MSG_RemotePrivate, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *name, char *msg)
{
	String s;
	s += ":";
	s += name;
	s += ":";
	s += msg;

	sendRemotePrivate(s);
}
